@@ -116,7 +116,7 @@ extern struct dico_stat current_stat, total_stat;
 
 #define DICTD_LOGGING_ENVAR "__DICTD_LOGGING__"
 
-#define MODE_DAEMON  0 
+#define MODE_DAEMON  0
 #define MODE_INETD   1
 #define MODE_PREPROC 2
 #define MODE_TEST    3
@@ -148,33 +148,37 @@ extern dicod_acl_t connect_acl;
 typedef struct dicod_module_instance {
     char *ident;
     char *command;
-    struct dico_database_module *module; 
+    struct dico_database_module *module;
     lt_dlhandle handle;
 } dicod_module_instance_t;
 
-#define DICOD_DBF_LANG 0x0001
+#define DICO_DBF_LANG    0x10000
 
 typedef struct dicod_database {
     int flags;
-    
+
     char *name;        /* Dictionary name */
     char *descr;       /* Description (SHOW DB) */
     char *info;        /* Info (SHOW INFO) */
-    
+
     dico_list_t langlist[2]; /* List of "source/dest" languages */
-    
+
     dicod_acl_t  acl;  /* ACL for this database */
-    int visible;       /* Result of the last dicod_acl_check */
-    
+    int visible;         /* Is this database (administratively) visible */
+    int session_visible; /* Is this database visible in the current session.
+			    This depends on the value of visible and on the
+			    result of dicod_acl_check and dicod_lang_check */
+
     dico_handle_t mod_handle;        /* Dico module handle */
 
     dico_assoc_list_t mime_headers;
-    
+
     dicod_module_instance_t *instance; /* Pointer to the module instance
 					  structure */
     int argc;                 /* Handler arguments: count */
     char **argv;              /*  ... and pointers */
     char *command;            /* Handler command line (for diagnostics) */
+    void *extra;
 } dicod_database_t;
 
 #define CONTENT_TRANSFER_ENCODING_HEADER "Content-transfer-encoding"
@@ -193,7 +197,6 @@ int get_input_line(dico_stream_t str, char **buf, size_t *size,
 
 dicod_database_t *find_database(const char *name);
 void database_remove_dependent(dicod_module_instance_t *inst);
-void dicod_database_free(dicod_database_t *dp);
 size_t database_count(void);
 int database_iterate(dico_list_iterator_t fun, void *data);
 int show_sys_info_p(void);
@@ -203,7 +206,16 @@ void dicod_log_encode_envar(void);
 char *get_full_hostname(void);
 void check_db_visibility(void);
 void reset_db_visibility(void);
-#define database_visible_p(db) ((db)->visible)
+static inline int
+database_is_visible(dicod_database_t const *db)
+{
+    return db->session_visible;
+}
+static inline int
+database_is_virtual(dicod_database_t const *db)
+{
+    return db->flags & DICO_DBF_VIRTUAL;
+}
 int dicod_any_lang_list_p(dico_list_t list);
 
 
@@ -237,6 +249,7 @@ void register_lang(void);
 int dicod_lang_check(dico_list_t list[2]);
 
 /* mime.c */
+extern int option_mime;
 void register_mime(void);
 
 /* markup.c */
@@ -266,18 +279,7 @@ void init_auth_data(void);
 /* loader.c */
 void dicod_loader_init(void);
 int dicod_load_module(dicod_module_instance_t *hptr);
-int dicod_init_database(dicod_database_t *dp);
-int dicod_open_database(dicod_database_t *dp);
-int dicod_close_database(dicod_database_t *dp);
-int dicod_free_database(dicod_database_t *dp);
 
-int dicod_database_get_strats(dicod_database_t *dp);
-
-char *dicod_get_database_descr(dicod_database_t *db);
-void dicod_free_database_descr(dicod_database_t *db, char *descr);
-char *dicod_get_database_info(dicod_database_t *db);
-void dicod_free_database_info(dicod_database_t *db, char *info);
-void dicod_get_database_languages(dicod_database_t *db, dico_list_t list[]);
 dico_list_t dicod_langlist_copy(dico_list_t src);
 
 void dicod_match_word_db(dicod_database_t *db, dico_stream_t stream,
@@ -293,6 +295,43 @@ void dicod_define_word_all(dico_stream_t stream, const char *word);
 
 int dicod_module_test(int argc, char **argv);
 
+void dicod_builtin_module_init(void);
+
+/* database.c */
+int dicod_database_init(dicod_database_t *dp);
+int dicod_database_open(dicod_database_t *dp);
+int dicod_database_close(dicod_database_t *dp);
+int dicod_database_deinit(dicod_database_t *dp);
+
+void dicod_database_free(dicod_database_t *dp);
+
+char *dicod_database_get_descr(dicod_database_t *db);
+void dicod_database_free_descr(dicod_database_t *db, char *descr);
+char *dicod_database_get_info(dicod_database_t *db);
+void dicod_database_free_info(dicod_database_t *db, char *info);
+void dicod_database_get_languages(dicod_database_t *db, dico_list_t list[]);
+
+size_t dicod_database_result_count(dicod_database_t *db, dico_result_t res);
+size_t dicod_database_compare_count(dicod_database_t *db, dico_result_t res);
+void dicod_database_result_free(dicod_database_t *db, dico_result_t res);
+int dicod_database_result_output(dicod_database_t *db, dico_result_t res,
+				 size_t n, dico_stream_t str);
+
+dico_result_t dicod_database_match(dicod_database_t *db,
+				   const dico_strategy_t strat,
+				   const char *word);
+dico_result_t dicod_database_define(dicod_database_t *db, const char *word);
+
+char *dicod_database_get_info(dicod_database_t *db);
+void dicod_database_free_info(dicod_database_t *db, char *info);
+
+char *dicod_database_get_descr(dicod_database_t *db);
+void dicod_database_free_descr(dicod_database_t *db, char *descr);
+
+dico_assoc_list_t dicod_database_mime_header(dicod_database_t *db,
+					     dico_result_t res);
+int dicod_database_flags(dicod_database_t const *db);
+
 /* ostream.c */
 extern off_t total_bytes_out;
 dico_stream_t dicod_ostream_create(dico_stream_t str,
@@ -370,4 +409,22 @@ int dicod_check_password(const char *db_pass, const char *pass);
 
 /* main.c */
 void config_help(void);
+int database_session_visibility(dicod_database_t *db);
 
+
+/* virtual.c */
+extern struct dico_database_module virtual_builtin_module;
+
+enum mime_cond {
+    cond_any = -1,
+    cond_mime,
+    cond_nomime
+};
+
+struct vdb_member {
+    char *name;
+    dicod_database_t *db;
+    int   cond;
+};
+
+dico_list_t vdb_list_create(void);
